/*
    AnaCam.  Take anaglyph pictures on Android.
    Copyright (C) 2012  Daniel Kraft <d@domob.eu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Native JNI implementation of anaglyph creation for speed.  */

#include "eu_domob_anacam_Anaglyph.h"
#include "eu_domob_anacam_MyBitmap.h"
#include "eu_domob_anacam_YuvDecoder.h"

#include <cassert>
#include <cstddef>
#include <cstdint>
#include <cstdio>
#include <cstring>

#include "jpeglib.h"

//#undef NDEBUG

#undef assert
#ifdef NDEBUG
  #define assert(cond)
#else /* NDEBUG.  */
  #define assert(cond) \
    if (!(cond)) \
      { \
        std::FILE* f = std::fopen ("/sdcard/anacam.log", "a"); \
        std::fprintf (f, "Assert failed: %s\n", #cond); \
        std::fclose (f); \
      }
#endif /* NDEBUG.  */

/* ************************************************************************** */
/* Anaglyph.  */

/**
 * Helper routine that does multiply-blending.
 * @param a First pixel value.
 * @param b Second pixel value.
 * @return Value of both multiplied together.
 */
static jint
multiply (jint a, jint b)
{
  jint res = 0xFF000000;
  for (unsigned channel = 0; channel < 3; ++channel)
    {
      const unsigned shift = 8 * channel;
      const jint mask = (0xFF << shift);
      const jint compA = ((a & mask) >> shift);
      const jint compB = ((b & mask) >> shift);
      assert (compA >= 0 && compA <= 255);
      assert (compB >= 0 && compB <= 255);
      const jint compRes = compA * compB / 0xFF;
      assert ((compRes & ~0xFF) == 0);
      res |= (compRes << shift);
    }

  return res;
}

/**
 * Helper routine to implement anaglyph combination in a common way
 * between the two versions.
 * @param len Number of pixels to process.
 * @param lcol Colour for left image blending.
 * @param rcol Colour for right image blending.
 * @param limg Native data pointer for left image.
 * @param rimg Native data pointer for right image.
 * @param dpx Array to hold the result pixels.
 */
static void
combineAnaglyph (const jsize len, jint lcol, jint rcol,
                 jlong limg, jlong rimg, jint* dpx)
{
  const jint* lpx = reinterpret_cast<const jint*> (limg);
  const jint* rpx = reinterpret_cast<const jint*> (rimg);

  /* Check the assumption made below that the screening colours
     only have 0 and 255 as intensities.  */
#ifndef NDEBUG
  for (unsigned channel = 0; channel < 4; ++channel)
    {
      const unsigned shift = 8 * channel;
      const uint32_t mask = (0xFFu << shift);
      const uint32_t l = ((lcol & mask) >> shift);
      const uint32_t r = ((rcol & mask) >> shift);
      assert (l == 0xFFu || l == 0x00u);
      assert (r == 0xFFu || r == 0x00u);
    }
#endif /* NDEBUG.  */

  /* We assume that the screening colours have only 0 and 255 intensities.
     This is checked above.  Then the screening can be expressed without
     multiplication, simply as bitwise OR.

     If the colours are *exactly* complementary (like red-cyan), then
     we can further simplify the calculation:  Then for each channel
     always one of the screened pixels stays untouched (where screening
     colour is 0), and the other gets set to 255 (where
     screening colour is 255).  Then their product is simply the value
     of the untouched pixel.  */
  const bool complementary = ((rcol & ~0xFF000000) == ~lcol);

  if (complementary)
    for (jsize i = 0; i < len; ++i)
      dpx[i] = ((lpx[i] & ~lcol) | (rpx[i] & ~rcol) | 0xFF000000);
  else
    for (jsize i = 0; i < len; ++i)
      {
        const jint lscr = (lpx[i] | lcol);
        const jint rscr = (rpx[i] | rcol);
        dpx[i] = multiply (lscr, rscr);
      }
}

JNIEXPORT void JNICALL
Java_eu_domob_anacam_Anaglyph_doCombine__IIJJ_3I (JNIEnv* env, jclass clazz,
                                                  jint lcol, jint rcol,
                                                  jlong limg, jlong rimg,
                                                  jintArray dest)
{
  const jsize len = env->GetArrayLength (dest);
  jint* dpx = env->GetIntArrayElements (dest, NULL);

  combineAnaglyph (len, lcol, rcol, limg, rimg, dpx);

  env->ReleaseIntArrayElements (dest, dpx, 0);
}

JNIEXPORT jlong JNICALL
Java_eu_domob_anacam_Anaglyph_doCombine__IIJJI (JNIEnv* env, jclass clazz,
                                                jint lcol, jint rcol,
                                                jlong limg, jlong rimg,
                                                jint len)
{
  jint* dpx = new jint[len];
  combineAnaglyph (len, lcol, rcol, limg, rimg, dpx);

  return reinterpret_cast<jlong> (dpx);
}

/* ************************************************************************** */
/* MyBitmap.  */

/* Helper routine to do grayscale in-place conversion.  */
static void
doGrayscale (const jsize len, jint* ptr)
{
  /* Use relative luminances of the colour channels from:
     http://stackoverflow.com/questions/8025621/how-to-convert-rgb565-to-yuv420sp-faster-on-android.  */
  static const jint Y[] = {19595, 38470, 7471};
  static const unsigned SHIFT = 16;

  for (jsize i = 0; i < len; ++i)
    {
      const jint cur = ptr[i];
      jint res = 0;
      for (unsigned channel = 0; channel < 3; ++channel)
        {
          const unsigned shift = 8 * channel;
          const jint mask = (0xFF << shift);
          const jint comp = ((cur & mask) >> shift);
          assert (comp >= 0 && comp <= 255);
          res += comp * Y[channel];
        }
      res >>= SHIFT;
      assert (res >= 0 && res <= 255);

      jint combined = 0xFF000000;
      for (unsigned channel = 0; channel < 3; ++channel)
        {
          const unsigned shift = 8 * channel;
          combined |= (res << shift);
        }

      ptr[i] = combined;
    }
}

JNIEXPORT void JNICALL
Java_eu_domob_anacam_MyBitmap_freeNative (JNIEnv* env, jclass clazz, jlong ptr)
{
  jint* cPtr = reinterpret_cast<jint*> (ptr);
  assert (cPtr);
  delete[] cPtr;
}

JNIEXPORT jlong JNICALL
Java_eu_domob_anacam_MyBitmap_copyToNative (JNIEnv* env, jclass clazz,
                                            jintArray data)
{
  const jsize len = env->GetArrayLength (data);
  jint* jPtr = env->GetIntArrayElements (data, NULL);

  jint* cPtr = new jint[len];
  std::memcpy (cPtr, jPtr, sizeof (jint) * len);

  env->ReleaseIntArrayElements (data, jPtr, 0);

  return reinterpret_cast<jlong> (cPtr);
}

JNIEXPORT void JNICALL
Java_eu_domob_anacam_MyBitmap_doGrayscale___3I (JNIEnv* env, jclass clazz,
                                                jintArray data)
{
  const jsize len = env->GetArrayLength (data);
  jint* ptr = env->GetIntArrayElements (data, NULL);

  doGrayscale (len, ptr);

  env->ReleaseIntArrayElements (data, ptr, 0);
}

JNIEXPORT void JNICALL
Java_eu_domob_anacam_MyBitmap_doGrayscale__JI (JNIEnv* env, jclass clazz,
                                               jlong ptr, jint len)
{
  jint* realPtr = reinterpret_cast<jint*> (ptr);
  doGrayscale (len, realPtr);
}

JNIEXPORT void JNICALL
Java_eu_domob_anacam_MyBitmap_saveToJpeg (JNIEnv* env, jclass clazz,
                                          jlong ptr, jint w, jint h,
                                          jstring file, jint quality)
{
  /* This code is based on example.c of libjpeg.  */

  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;
  std::FILE* outfile;

  cinfo.err = jpeg_std_error (&jerr);
  jpeg_create_compress (&cinfo);

  const char* cFile = env->GetStringUTFChars (file, NULL);
  outfile = std::fopen (cFile, "wb");
  env->ReleaseStringUTFChars (file, cFile);
  jpeg_stdio_dest (&cinfo, outfile);

  cinfo.image_width = w;
  cinfo.image_height = h;
  cinfo.input_components = 3;
  cinfo.in_color_space = JCS_RGB;
  jpeg_set_defaults (&cinfo);
  jpeg_set_quality (&cinfo, quality, TRUE);

  jpeg_start_compress (&cinfo, TRUE);

  const jint* data = reinterpret_cast<const jint*> (ptr);
  JSAMPLE* row = new JSAMPLE[3 * w];
  JSAMPROW row_ptr[] = {row};
  while (cinfo.next_scanline < cinfo.image_height)
    {
      const jint* curData = data + w * cinfo.next_scanline;
      for (unsigned i = 0; i < w; ++i)
        {
          const jint cur = curData[i];
          assert ((cur & 0xFF000000) == 0xFF000000);
          JSAMPLE* out = row + 3 * i;
          out[0] = ((cur & 0x00FF0000) >> 2 * 8);
          out[1] = ((cur & 0x0000FF00) >> 1 * 8);
          out[2] = (cur & 0x000000FF);
        }
      jpeg_write_scanlines (&cinfo, row_ptr, 1);
    }
  delete[] row;

  jpeg_finish_compress (&cinfo);
  std::fclose (outfile);

  jpeg_destroy_compress (&cinfo);
}

JNIEXPORT void JNICALL
Java_eu_domob_anacam_MyBitmap_loadJpeg (JNIEnv* env, jobject thiz,
                                        jbyteArray jpeg)
{
  /* This code is based on example.c of libjpeg.  */

  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;

  cinfo.err = jpeg_std_error (&jerr);
  jpeg_create_decompress (&cinfo);

  const jsize len = env->GetArrayLength (jpeg);
  jbyte* cJpeg = env->GetByteArrayElements (jpeg, NULL);
  unsigned char* inBuf = reinterpret_cast<unsigned char*> (cJpeg);
  jpeg_mem_src (&cinfo, inBuf, len);

  jpeg_read_header (&cinfo, TRUE);
  jpeg_start_decompress (&cinfo);

  const jint w = cinfo.output_width;
  const jint h = cinfo.output_height;

  assert (cinfo.output_components == 3);
  assert (cinfo.out_color_space == JCS_RGB);
  JSAMPARRAY buffer = (*cinfo.mem->alloc_sarray)
                        (reinterpret_cast<j_common_ptr> (&cinfo),
                         JPOOL_IMAGE, 3 * w, 1);

  jint* data = new jint[w * h];
  assert (data);
  while (cinfo.output_scanline < cinfo.output_height)
    {
      /* Do this with the old value of output_scanline, thus we have
         to have this line before jpeg_read_scanlines.  */
      jint* curData = data + w * cinfo.output_scanline;

      jpeg_read_scanlines (&cinfo, buffer, 1);

      const JSAMPLE* row = buffer[0];
      for (unsigned i = 0; i < w; ++i)
        {
          const JSAMPLE* in = row + 3 * i;
          jint cur = 0xFF000000;
          cur |= (in[0] << 2 * 8);
          cur |= (in[1] << 1 * 8);
          cur |= (in[2]);
          curData[i] = cur;
        }
    }

  jpeg_finish_decompress (&cinfo);
  jpeg_destroy_decompress (&cinfo);
  env->ReleaseByteArrayElements (jpeg, cJpeg, 0);

  jclass clazz = env->GetObjectClass (thiz);
  jfieldID field;

  field = env->GetFieldID (clazz, "width", "I");
  env->SetIntField (thiz, field, w);
  field = env->GetFieldID (clazz, "height", "I");
  env->SetIntField (thiz, field, h);

  field = env->GetFieldID (clazz, "nativeData", "J");
  env->SetLongField (thiz, field, reinterpret_cast<jlong> (data));
}

/* ************************************************************************** */
/* YuvDecoder.  */

JNIEXPORT void JNICALL
Java_eu_domob_anacam_YuvDecoder_decodeYUV420SP
  (JNIEnv* env, jclass clazz, jintArray javaRGB, jbyteArray javaYUV,
   jint width, jint height, jboolean gray)
{
  /* This code is from:
     https://code.google.com/p/android/issues/detail?id=22966.  */

  jint* rgb = env->GetIntArrayElements (javaRGB, NULL);
  jbyte* yuv = env->GetByteArrayElements (javaYUV, NULL);

  const jint frameSize = width * height;
  for (jint j = 0, yp = 0; j < height; j++)
    {
      jint uvp = frameSize + (j >> 1) * width;
      jint u = 0, v = 0;
      for (jint i = 0; i < width; i++, yp++)
        {
          jint y = (0xFF & yuv[yp]) - 16;
          if (y < 0)
            y = 0;
          assert (y >= 0 && y <= 255);

          if (gray)
            {
              jint res = 0xFF000000;
              for (unsigned channel = 0; channel < 3; ++channel)
                {
                  const unsigned shift = 8 * channel;
                  res |= (y << shift);
                }
              rgb[yp] = res;
            }
          else
            {
              if ((i & 1) == 0)
                {
                  v = (0xFF & yuv[uvp++]) - 128;
                  u = (0xFF & yuv[uvp++]) - 128;
                }

              const jint y1192 = 1192 * y;
              jint r = (y1192 + 1634 * v);
              jint g = (y1192 - 833 * v - 400 * u);
              jint b = (y1192 + 2066 * u);
           
              if (r < 0)
                r = 0;
              else if (r > 262143)
                r = 262143;

              if (g < 0)
                g = 0;
              else if (g > 262143)
                g = 262143;

              if (b < 0)
                b = 0;
              else if (b > 262143)
                b = 262143;

              rgb[yp] = 0xFF000000 | ((r << 6) & 0xFF0000)
                                   | ((g >> 2) & 0xFF00) | ((b >> 10) & 0xFF);
            }
        } 
    }

  env->ReleaseIntArrayElements (javaRGB, rgb, 0);
  env->ReleaseByteArrayElements (javaYUV, yuv, 0);
}
