/*
    AnaCam.  Take anaglyph pictures on Android.
    Copyright (C) 2012  Daniel Kraft <d@domob.eu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package eu.domob.anacam;

import android.graphics.ImageFormat;

/**
 * Decode YUV camera data to RGB.  Of course there's also
 * android.graphics.YuvImage, but I don't want the detour with compressing
 * and decompressing a jpeg.  This really, really shoudn't be necessary....
 *
 * Note that this class internally keeps the RGB array allocated and shares
 * it between decodes.  This is ok since decoding is only ever done for
 * the preview frames, and it should help reduce memory overhead.
 */
public class YuvDecoder
{

  /** Byte YUV data.  */
  private byte[] yuv;

  /** Buffer to hold RGB data.  */
  private int[] rgb;

  /** Width of the image and allocated RGB buffer.  */
  private int width;
  /** Height of the image and allocated RGB buffer.  */
  private int height;

  /** Format of the YUV data.  */
  private int format;

  /**
   * Construct it empty.
   */
  public YuvDecoder ()
  {
    yuv = null;
    rgb = null;
  }

  /**
   * Give it YUV data.  This (re)allocates the RGB buffer as necessary.
   * @param px YUV data.
   * @param w Image width.
   * @param h Image height.
   * @param fmt Image format.
   */
  public void setInput (byte[] px, int w, int h, int fmt)
  {
    if (rgb == null || w != width || h != height)
      {
        rgb = new int[w * h];
        width = w;
        height = h;
      }

    yuv = px;
    format = fmt;
  }

  /**
   * Free buffers to make memory available.
   */
  public void freeBuffers ()
  {
    rgb = null;
  }

  /**
   * Decode it to a Bitmap in RGB format.
   * @param gray Decode as grayscale?
   * @return The constructed bitmap.
   */
  public MyBitmap decode (boolean gray)
  {
    if (format != ImageFormat.NV21)
      throw new RuntimeException ("Only support NV21 decoding for now.");

    decodeYUV420SP (rgb, yuv, width, height, gray);
    yuv = null;

    return new MyBitmap (rgb, width, height);
  }

  /**
   * Perform the actual decoding.
   * @param rgb Store RGB pixels here.
   * @param yuv YUV data to decode.
   * @param width Width of the image.
   * @param height Height of the image.
   * @param gray Decode as grayscale?
   */
  private static native void decodeYUV420SP (int[] rgb, byte[] yuv, int width,
                                             int height, boolean gray);

  static
    {
      System.loadLibrary ("anacam");
    }

}
